"""Exceptions triggered to warn about internal errors related to Bluesky DataParser Factories classes."""


class BlueskyDataParserFactoryError(Exception):
    """Custom Exception to warn when any error occurs inside Bluesky DataParser Factories."""
