"""Exceptions triggered to warn about internal errors related to Assonant data logger classes."""


class AssonantDataLoggerError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantDataLogger."""
