"""Exceptions triggered to warn about internal errors related to Assonant metadata retriever classes."""


class AssonantMetadataRetrieverError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantMetadataRetriever."""
