"""Assonant Data Logging module.

Assonant toolkit for automatically logging data based on data mapping/configration files.
"""

from .assonant_data_logger import AssonantDataLogger

__all__ = [
    "AssonantDataLogger",
]
