"""Assonant Beam Visualiation System (BVS) component data class."""
from .component import Component


class BVS(Component):
    """Data class to handle all data required to define a Beam Visualization System (BVS)."""
