"""Exceptions triggered to warn about internal errors related to Bluesky DataParser class."""


class BlueskyDataParserError(Exception):
    """Custom Exception to warn when any error occurs inside Bluesky DataParser."""
