"""Assonant data logger exceptions.

This submodule defines Exceptions used all over the data logger module and its submodules.
"""

from .data_loggers_exceptions import AssonantDataLoggerError

__all__ = ["AssonantDataLoggerError"]
