"""Assonant Bending Magnet component data class."""

from .component import Component


class BendingMagnet(Component):
    """Data class to handle all data required to define a bending magnet insertion device."""
