"""Assonant Granite base component data class."""

from .component import Component


class GraniteBase(Component):
    """Data class to handle all data required to define granite base where many components are mounted over it."""
