"""Assonant metadata retriever.

This submodule defines Metadata Retrievers objects to allow retrieving important
metadata for Assonant from different sources.
"""

from .assonant_metadata_retriever import AssonantMetadataRetriever

__all__ = [
    "AssonantMetadataRetriever",
]
