"""Assonant data retriever exceptions.

This submodule defines Exceptions used all over the data retriever module and its submodules.
"""

from .data_retrievers_exceptions import AssonantDataRetrieverError

__all__ = ["AssonantDataRetrieverError"]
