"""Assonant Storage Ring component data class."""

from .component import Component


class StorageRing(Component):
    """Data class to handle all data required to define a particle accelerator storage ring."""
