"""Assonant Fresnel Zone Plate component data class."""

from .component import Component


class FresnelZonePlate(Component):
    """Data class to handle all data required to define a fresnel zone plate."""
